import { unzip } from 'unzipit';
import { downloadZip } from 'client-zip';
import { RDFMimetype, AsyncOxigraph } from 'async-oxigraph';
import * as i0 from '@angular/core';
import { Injectable, EventEmitter, Component, Output, HostListener, Input, ViewChild, NgModule } from '@angular/core';
import { GLTFExporter } from 'three/examples/jsm/exporters/GLTFExporter';
import { IfcFragmentLoader, Components, SimpleScene, PostproductionRenderer, OrthoPerspectiveCamera, SimpleRaycaster, EdgesClipper, EdgesPlane, SimpleDimensions, SimpleGrid, Fragments } from 'openbim-components';
import { v5 } from 'uuid';
import * as SparkMD5 from 'spark-md5';
import { IfcAPI } from 'web-ifc';
import { SerializationFormat, LBDParser, ParserSettings } from 'ifc-lbd';
import { BehaviorSubject, debounceTime, fromEvent } from 'rxjs';
import * as i4 from '@angular/common';
import { CommonModule } from '@angular/common';
import * as i5 from '@angular/forms';
import { FormsModule } from '@angular/forms';
import { HttpClientModule } from '@angular/common/http';
import * as i1 from '@angular/material/icon';
import { MatIconModule } from '@angular/material/icon';
import { MatLegacyButtonModule } from '@angular/material/legacy-button';
import * as i7 from '@angular/material/legacy-tooltip';
import { MatLegacyTooltipModule } from '@angular/material/legacy-tooltip';
import { DirectionalLight, AmbientLight, MeshBasicMaterial, Matrix4 } from 'three';
import * as i2 from '@angular/platform-browser';

// Trick used by FileSaver.js:
// 1. Create a hidden <a> tag.
// 2. Set its href attribute to the blob's URL.
// 3. Set its download attribute to the filename.
// 4. Click on the <a> tag.
const saveBlob = (function () {
    let a = document.createElement("a");
    document.body.appendChild(a);
    // a.style = "display: none";
    return function (blob, fileName) {
        const url = window.URL.createObjectURL(blob);
        a.href = url;
        a.download = fileName;
        a.click();
        window.URL.revokeObjectURL(url);
        return url;
    };
}());

class ICDD {
    constructor() {
        this.entries = [];
        this.fileName = "";
    }
    async loadContainer(icdd, workerPath = "./assets/oxigraph/worker.js", oxigraphWASMPathRelativeToWorker = "./") {
        this.fileName = icdd.name;
        const fileURL = URL.createObjectURL(icdd);
        const zipInfo = await unzip(fileURL);
        this.entries = Object.values(zipInfo.entries);
        await this.initStore(workerPath, oxigraphWASMPathRelativeToWorker);
    }
    // Query (responseMimetype only used for construct)
    async query(query, responseMimetype) {
        if (this.asyncOxigraph == undefined) {
            console.log("Please initialize Oxigraph before executing queries");
            return;
        }
        const queryResponse = await this.asyncOxigraph.query(query, responseMimetype);
        return queryResponse.data;
    }
    async load(triples, mimetype, baseURI, graphURI) {
        if (this.asyncOxigraph == undefined) {
            const msg = "Please initialize Oxigraph before executing queries";
            console.log(msg);
            return msg;
        }
        const loadResponse = await this.asyncOxigraph.load(triples, mimetype, baseURI, graphURI);
        return loadResponse.message;
    }
    async getNQuadsFiles() {
        let textPromises = [];
        for (let i = 0; i < this.entries.length; i++) {
            const file = this.entries[i];
            if (file.name.endsWith(".nq")) {
                textPromises.push(file.text());
            }
        }
        return await Promise.all(textPromises);
    }
    // Close ICDD to free up memory
    async close(downloadICDD = false) {
        if (downloadICDD)
            await this.download();
        this.entries = [];
        this.asyncOxigraph?.close();
    }
    async download() {
        if (this.asyncOxigraph == undefined) {
            const msg = "Nothing to download. Please initialize Oxigraph.";
            console.log(msg);
            return;
        }
        // // Write all other files than the nq
        let filePromises = [];
        this.entries.forEach(entry => {
            if (entry.name != "Payload triples/data.nq") {
                filePromises.push(this.entryToFile(entry));
            }
        });
        let files = await Promise.all(filePromises);
        // Write nquads
        const res = await this.asyncOxigraph.dump(RDFMimetype.NQUADS);
        files.push(new File([res.data], "Payload triples/data.nq"));
        // get the zip file as a Blob
        const blob = await downloadZip(files).blob();
        saveBlob(blob, this.fileName);
    }
    async initStore(workerPath = "./assets/oxigraph/worker.js", oxigraphWASMPathRelativeToWorker = "./") {
        // Init Oxigraph
        this.asyncOxigraph = new AsyncOxigraph(workerPath);
        await this.asyncOxigraph.init();
        // Get NQuads file from zip
        const nquads = await this.getNQuadsFiles();
        // Load data in background
        for (let nq of nquads) {
            const res = await this.asyncOxigraph.load(nq, RDFMimetype.NQUADS, undefined, undefined);
            console.log(res.message);
        }
    }
    async entryToFile(entry) {
        const blob = await entry.blob();
        return new File([blob], entry.name);
    }
}

class ICDDWriter {
    constructor() {
        this.fileName = "model.icdd";
        this.files = [];
        // private async serializeStoreContent(){
        //     // Serialize index.rdf
        //     let triples = await this.store?.dump("application/n-triples", oxigraph.namedNode("https://index")) as string;
        //     this.files.push(new File([triples], `index.rdf`));
        //     // Serialize main content
        //     triples = await this.store?.dump("application/n-triples", undefined) as string;
        //     this.files.push(new File([triples], `Payload triples/data.nq`));
        //     triples = "";
        //     // Close store
        //     // await this.store?.free();
        //     this.store = undefined;
        // }
    }
    async init(ifcFileName) {
        const ext = ifcFileName.split(".").pop();
        this.fileName = ext == undefined ? ifcFileName + ".ifc" : ifcFileName.replace(ext, "zip");
    }
    async close() {
        // get the zip file as a Blob
        const blob = await downloadZip(this.files).blob();
        saveBlob(blob, this.fileName);
    }
    async addPayloadTriples(fileName, triples) {
        const filePath = `Payload triples/${fileName}`;
        // Write to zip
        this.files.push(new File(triples, filePath));
    }
    async addJSONDocument(fileName, json) {
        const filePath = `Payload documents/${fileName}`;
        this.files.push(new File([JSON.stringify(json)], filePath));
    }
    async addPayloadDocument(fileName, blobs) {
        const filePath = `Payload documents/${fileName}`;
        const fileURI = `https://Payload_documents/${encodeURIComponent(fileName)}`;
        // Write to zip
        this.files.push(new File(blobs, filePath));
        // // Add to index
        // this.store?.add(oxigraph.quad(
        //     oxigraph.namedNode(fileURI),
        //     oxigraph.namedNode("http://www.w3.org/1999/02/22-rdf-syntax-ns#"),
        //     oxigraph.namedNode("https://standards.iso.org/iso/21598/-1/ed-1/en/Container#Document"),
        //     oxigraph.namedNode("https://index")
        // ));
        // this.store?.add(oxigraph.quad(
        //     oxigraph.namedNode(fileURI),
        //     oxigraph.namedNode("https://standards.iso.org/iso/21598/-1/ed-1/en/Container#filePath"),
        //     oxigraph.literal(filePath, oxigraph.namedNode("http://www.w3.org/2001/XMLSchema#anyURI")),
        //     oxigraph.namedNode("https://index")
        // ));
        // this.store?.add(oxigraph.quad(
        //     oxigraph.namedNode(fileURI),
        //     oxigraph.namedNode("https://standards.iso.org/iso/21598/-1/ed-1/en/Container#creationDate"),
        //     oxigraph.literal(new Date().toISOString(), oxigraph.namedNode("http://www.w3.org/2001/XMLSchema#dateTime")),
        //     oxigraph.namedNode("https://index")
        // ));
    }
}

async function contextBasedGuid(contextString) {
    const namespace = 'daca0510-72b5-48ba-9091-b918ca18136b';
    return v5(contextString, namespace);
}
async function md5(file) {
    return new Promise((resolve) => {
        const spark = new SparkMD5.ArrayBuffer();
        spark.append(file);
        const hash = spark.end();
        resolve(hash);
    });
}

async function readFile(file) {
    return new Promise((resolve, reject) => {
        // Create file reader
        let reader = new FileReader();
        // Register event listeners
        reader.addEventListener("loadend", (e) => resolve(e.target.result));
        reader.addEventListener("error", reject);
        // Read file
        reader.readAsArrayBuffer(file);
    });
}

class ParsingProgress {
    constructor(modelLoad, rdfParse, zipWrite) {
        this.modelLoad = 0;
        this.rdfParse = 0;
        this.zipWrite = 0;
        this.modelLoad = modelLoad;
        this.rdfParse = rdfParse;
        this.zipWrite = zipWrite;
    }
}
class SerializerService {
    constructor() {
        this.exporter = new GLTFExporter();
        this.icddWriter = new ICDDWriter();
        this.expressFragmentMap = {}; // Maps express id with geometry UUID
        this.expressGlobalMap = {}; // Maps express id with globalId
        this.parsingProgress = new ParsingProgress(0, 0, 0);
        this.parsingProgress$ = new BehaviorSubject(this.parsingProgress);
        this.webIFCWasmPath = "assets/ifcjs/";
    }
    trackProgress() {
        return this.parsingProgress$.asObservable();
    }
    async serailizeICDD(ifcFile, settings) {
        await this.init();
        const sizeInMB = (ifcFile.size / (1024 * 1024)).toFixed(2);
        console.log(`${ifcFile.name} | ${sizeInMB}MB`);
        const [model, triples] = await Promise.all([
            this.processIFCGeometry(ifcFile),
            this.parseRDF(ifcFile, settings)
        ]);
        // Init the ICDD
        await this.icddWriter.init(ifcFile.name);
        // Write RDF
        await this.icddWriter.addPayloadTriples("data.nq", [triples]);
        const expressIDs = Object.keys(model.expressIDFragmentIDMap).map((id) => parseInt(id));
        await Promise.all([
            this.writeProperties(model),
            this.writeFragments(model),
            this.buildExpressIdToGlobalIdMap(expressIDs)
        ]);
        // Write maps
        await this.icddWriter.addJSONDocument("express-fragment-map.json", this.expressFragmentMap);
        await this.icddWriter.addJSONDocument("express-global-map.json", this.expressGlobalMap);
        // Close the ICDD (downloads zip)
        await this.icddWriter.close();
        this.ifcAPI?.CloseModel(0);
        this.parsingProgress = new ParsingProgress(100, 100, 100);
        this.parsingProgress$.next(this.parsingProgress);
    }
    async init() {
        if (this.ifcAPI != undefined)
            this.ifcAPI.CloseModel(0);
        this.ifcAPI = new IfcAPI();
        this.ifcAPI.SetWasmPath(this.webIFCWasmPath);
        await this.ifcAPI.Init();
    }
    async processIFCGeometry(file) {
        const loader = new IfcFragmentLoader();
        loader.settings.webIfc = {
            COORDINATE_TO_ORIGIN: false,
            USE_FAST_BOOLS: true
        };
        loader.settings.wasm.path = this.webIFCWasmPath;
        loader.progress.on((event) => {
            const fileProcessing = Math.floor((event.current * 100) / event.total);
            this.parsingProgress.modelLoad = fileProcessing;
            this.parsingProgress$.next(this.parsingProgress);
        });
        // Parse the model and remove the loaded file
        const fileURL = URL.createObjectURL(file);
        const model = await loader.load(new URL(fileURL));
        this.parsingProgress.modelLoad = 100;
        this.parsingProgress$.next(this.parsingProgress);
        return model;
    }
    async parseRDF(file, settings) {
        if (this.ifcAPI == undefined)
            return "";
        const arrayBuffer = await readFile(file);
        const modelID = this.ifcAPI.OpenModel(new Uint8Array(arrayBuffer));
        settings.outputFormat = SerializationFormat.NQuads;
        const lbdParser = new LBDParser(settings);
        const nquads = await lbdParser.parse(this.ifcAPI, modelID);
        this.parsingProgress.rdfParse = 100;
        this.parsingProgress$.next(this.parsingProgress);
        return nquads;
    }
    async buildExpressIdToGlobalIdMap(expressIDs, modelID = 0) {
        if (this.ifcAPI == undefined)
            return;
        const map = {};
        for (const expressID of expressIDs) {
            const properties = await this.ifcAPI.properties.getItemProperties(modelID, expressID);
            const globalId = properties.GlobalId.value;
            map[expressID] = globalId;
        }
        this.ifcAPI.CloseModel(modelID);
        this.expressGlobalMap = map;
    }
    // Builds map between expressID and globalID
    async writeProperties(model) {
        // Write various relationships
        await this.icddWriter.addJSONDocument("level-relationships.json", model.levelRelationships);
        await this.icddWriter.addJSONDocument("item-types.json", model.itemTypes);
        await this.icddWriter.addJSONDocument("all-types.json", model.allTypes);
        await this.icddWriter.addJSONDocument("floors-properties.json", model.floorsProperties);
        await this.icddWriter.addJSONDocument("coordination-matrix.json", model.coordinationMatrix);
        await this.icddWriter.addJSONDocument("bounding-boxes-opaque.json", model.boundingBoxes);
        await this.icddWriter.addJSONDocument("bounding-boxes-transparent.json", model.transparentBoundingBoxes);
    }
    // Serialize glTF and data from fragments and append to the zip archive
    async writeFragments(model) {
        const fragments = model.fragments;
        const promises = [];
        fragments.forEach((fragment) => {
            promises.push(this.exportFragment(fragment));
        });
        await Promise.all(promises);
    }
    // Serialize glTF and data from single fragment and append to the zip archive
    async exportFragment(fragment) {
        // Convert fragment geometry to glTF
        // const geometryArrayBuffer: ArrayBuffer = await this.exportMesh(fragment.mesh);
        const geometryArrayBuffer = await fragment.mesh.export();
        // Create UUID that is unique for the geometry
        const hash = await md5(geometryArrayBuffer);
        const fileId = await contextBasedGuid(hash);
        // Write geometry
        const fileName = `${fileId}.glb`;
        await this.icddWriter.addPayloadDocument(fileName, [new Blob([geometryArrayBuffer])]);
        // Add to map
        fragment.items.forEach((expressId) => (this.expressFragmentMap[expressId] = fileId));
        // Write data
        const fragmentData = {
            matrices: Array.from(fragment.mesh.instanceMatrix.array),
            ids: fragment.items,
            id: fileId,
        };
        await this.icddWriter.addJSONDocument(`${fileId}.json`, fragmentData);
    }
}
SerializerService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: SerializerService, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
SerializerService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: SerializerService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: SerializerService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }] });

class LBDComponentsService {
    constructor(serializer) {
        this.serializer = serializer;
        this.settings = new ParserSettings();
    }
    trackLBDParsingProgress() {
        return this.serializer.trackProgress();
    }
    parseLBD(file, settings = new ParserSettings()) {
        this.settings = settings;
        this.settings.verbose = false;
        // Serialize results
        this.serializer.serailizeICDD(file, this.settings);
        return this.serializer.trackProgress();
    }
}
LBDComponentsService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: LBDComponentsService, deps: [{ token: SerializerService }], target: i0.ɵɵFactoryTarget.Injectable });
LBDComponentsService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: LBDComponentsService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: LBDComponentsService, decorators: [{
            type: Injectable,
            args: [{
                    providedIn: 'root'
                }]
        }], ctorParameters: function () { return [{ type: SerializerService }]; } });

class ViewerSettings {
    constructor() {
        this.highlightElementOnHover = true;
        this.highlightColor = 0xFFC300;
        this.selectionColor = 0xF3E1FF;
        this.selectionMode = SelectionMode.MULTIPLE;
        this.showPointerCursorOnElementHover = true;
        this.hideSpaces = true;
        this.refreshInterval = 300; // Debounce time in ms between new fragments loaded and update of views
        this.verbose = false;
    }
}
var SelectionMode;
(function (SelectionMode) {
    SelectionMode["DISABLED"] = "DISABLED";
    SelectionMode["SINGLE"] = "SINGLE";
    SelectionMode["MULTIPLE"] = "MULTIPLE";
})(SelectionMode || (SelectionMode = {}));
var Tool;
(function (Tool) {
    Tool["NONE"] = "NONE";
    Tool["DIMENSIONS"] = "DIMENSIONS";
    Tool["CLIPPER"] = "CLIPPER";
})(Tool || (Tool = {}));

class MenuSettings {
    constructor() {
        this.clipperMode = false;
        this.dimensionsMode = false;
        this.orbitMode = true;
        this.firstPersonMode = false;
        this.planMode = false;
        this.showGrid = false;
        this.showInfo = true;
        this.infoText = "";
    }
    setNavMode(mode) {
        this.disableAllNavModes();
        if (mode == "Plan")
            this.planMode = true;
        if (mode == "Orbit")
            this.orbitMode = true;
        if (mode == "FirstPerson")
            this.firstPersonMode = true;
    }
    toggleClipperMode(setting) {
        const currentSetting = this.clipperMode;
        this.disableAllTools();
        this.clipperMode = setting != undefined ? setting : !currentSetting;
    }
    toggleDimensionsMode(setting) {
        const currentSetting = this.dimensionsMode;
        this.disableAllTools();
        this.dimensionsMode = setting != undefined ? setting : !currentSetting;
    }
    toggleShowInfo() {
        this.showInfo = !this.showInfo;
    }
    toggleShowGrid() {
        this.showGrid = !this.showGrid;
    }
    disableAllTools() {
        this.clipperMode = false;
        this.dimensionsMode = false;
    }
    disableAllNavModes() {
        this.orbitMode = false;
        this.firstPersonMode = false;
        this.planMode = false;
    }
}
class VisibilityMenu {
    constructor() {
        this.visible = false;
        this.left = "0px";
        this.top = "0px";
        this.categories = [];
    }
}

async function buildFragmentMap(globalIds, idMap) {
    return new Promise(resolve => {
        let map = {};
        globalIds.forEach(globalId => {
            const elementData = idMap[globalId];
            if (elementData == undefined)
                return;
            if (Object.keys(map).indexOf(elementData.fragmentId) == -1) {
                map[elementData.fragmentId] = [elementData.expressId];
            }
            else {
                map[elementData.fragmentId].push(elementData.expressId);
            }
        });
        resolve(map);
    });
}
async function buildFragmentMapFromPicked(pickedElements) {
    return new Promise(resolve => {
        let map = {};
        pickedElements.forEach(el => {
            if (el.elementData != undefined) {
                if (Object.keys(map).indexOf(el.elementData.fragmentId) == -1) {
                    map[el.elementData.fragmentId] = [el.elementData.expressId];
                }
                else {
                    map[el.elementData.fragmentId].push(el.elementData.expressId);
                }
            }
        });
        resolve(map);
    });
}
async function buildFragmentModelMap(globalIds, idMap) {
    return new Promise(resolve => {
        let map = {};
        globalIds.forEach(globalId => {
            const elementData = idMap[globalId];
            if (elementData == undefined)
                return;
            if (Object.keys(map).indexOf(elementData.modelURL) == -1) {
                map[elementData.modelURL] = [elementData.fragmentId];
            }
            else {
                map[elementData.modelURL].push(elementData.fragmentId);
            }
        });
        resolve(map);
    });
}
// Takes a list of globalIds and returns the opposite (all other globalIds)
async function getAllRemainingGlobalIds(globalIds, idMap) {
    return new Promise(resolve => {
        let remainingGlobalIds = [];
        Object.values(idMap).forEach(item => {
            if (item.globalId != undefined && !globalIds.includes(item.globalId))
                remainingGlobalIds.push(item.globalId);
        });
        resolve(remainingGlobalIds);
    });
}
async function getAllOfType(types, idMap) {
    return new Promise(resolve => {
        let globalIds = [];
        Object.values(idMap).forEach(item => {
            if (item.type != undefined && types.includes(item.type))
                globalIds.push(item.globalId);
        });
        resolve(globalIds);
    });
}
async function getAllNotOfType(types, idMap) {
    return new Promise(resolve => {
        let globalIds = [];
        Object.values(idMap).forEach(item => {
            if (item.type != undefined && !types.includes(item.type))
                globalIds.push(item.globalId);
        });
        resolve(globalIds);
    });
}
async function getElementData(expressId, idMap) {
    return new Promise(resolve => {
        const match = Object.values(idMap).find(item => item.expressId == expressId);
        resolve(match);
    });
}
async function markAllUnpickable(idMap) {
    return new Promise(resolve => {
        Object.keys(idMap).forEach(key => idMap[key].unpickable = true);
        resolve();
    });
}
async function markAllPickable(idMap) {
    return new Promise(resolve => {
        Object.keys(idMap).forEach(key => idMap[key].unpickable = false);
        resolve();
    });
}
async function markPickable(globalIds, idMap) {
    return new Promise(resolve => {
        Object.keys(idMap).forEach(key => {
            if (globalIds.includes(idMap[key].globalId))
                idMap[key].unpickable = false;
        });
        resolve();
    });
}
async function markUnpickable(globalIds, idMap) {
    return new Promise(resolve => {
        Object.keys(idMap).forEach(key => {
            if (globalIds.includes(idMap[key].globalId))
                idMap[key].unpickable = true;
        });
        resolve();
    });
}

async function getZipContent(zipURL, modelIndex, modelName) {
    let zipContent = { elementIdMap: {}, fragmentTypeMap: {} };
    const { entries } = await unzip(zipURL);
    const fileNames = Object.keys(entries);
    const hasSlash = fileNames.find(fn => fn.indexOf("/") != -1);
    let folderPrefix = "Payload documents/"; // Default ICDD subfolder
    if (hasSlash == undefined)
        folderPrefix = "";
    zipContent.folderPrefix = folderPrefix;
    // Required content
    [`${folderPrefix}bounding-boxes-opaque.json`, `${folderPrefix}bounding-boxes-transparent.json`, `${folderPrefix}express-fragment-map.json`].forEach(item => {
        if (fileNames.indexOf(item) == -1)
            throw new Error(`Couldn't find a ${item} file in the zip!`);
    });
    zipContent.opaqueBoxes = await entries[`${folderPrefix}bounding-boxes-opaque.json`].json();
    zipContent.transparentBoxes = await entries[`${folderPrefix}bounding-boxes-transparent.json`].json();
    zipContent.fragmentMap = await entries[`${folderPrefix}express-fragment-map.json`].json();
    // Optional id map
    try {
        zipContent.expressGlobalMap = await entries[`${folderPrefix}express-global-map.json`].json();
    }
    catch (err) {
        console.warn("Zip doesn't contain am \"express-global-map.json\" and therefore it will not be possible to emit GlobalIds from objects");
    }
    // Optional types
    let allTypes;
    let expressTypeMap;
    try {
        allTypes = await entries[`${folderPrefix}all-types.json`].json();
        expressTypeMap = await entries[`${folderPrefix}item-types.json`].json();
    }
    catch (err) {
        console.warn("Zip doesn't contain an \"item-types.json\" and/or \"all-types.json\" and therefore it will not be possible to use type dependent functionalities");
    }
    // Optional floors
    let floorProperties = [];
    let levelRelationships;
    try {
        floorProperties = await entries[`${folderPrefix}floors-properties.json`].json();
        levelRelationships = await entries[`${folderPrefix}level-relationships.json`].json();
    }
    catch (err) {
        console.warn("Zip doesn't contain an \"floors-properties.json\" and/or \"level-relationships.json\" and therefore it will not be possible to use type dependent functionalities");
    }
    // Build element id map
    zipContent.idMap = {};
    const expressIds = Object.keys(zipContent.expressGlobalMap);
    const globalIds = Object.values(zipContent.expressGlobalMap);
    globalIds.forEach((globalId, i) => {
        const expressId = expressIds[i];
        const typeId = expressTypeMap[expressId] ?? undefined;
        const levelId = levelRelationships[expressId] ?? undefined;
        const fragmentId = zipContent.fragmentMap[expressId];
        const type = allTypes[typeId];
        let levelData;
        if (typeId != undefined && floorProperties.length) {
            const lvl = floorProperties.find((item) => item.expressID == levelId);
            if (lvl != undefined) {
                levelData = { expressId: levelId, globalId: lvl.GlobalId.value, name: lvl.Name.value, elevation: lvl.Elevation.value };
            }
        }
        zipContent.elementIdMap[globalId] = {
            expressId,
            globalId,
            modelIndex,
            modelName,
            modelURL: zipURL,
            fragmentId,
            typeId,
            type,
            levelId,
            levelData
        };
        zipContent.fragmentTypeMap[fragmentId] = type;
    });
    // Build model id map
    zipContent.fragmentIdToModelIdxMap = {};
    Object.values(zipContent.fragmentMap).forEach(uuid => {
        const fragmentId = uuid;
        zipContent.fragmentIdToModelIdxMap[fragmentId] = modelIndex;
    });
    return zipContent;
}

class BimViewerService {
    constructor() {
        this.sceneReady = new Promise((resolve) => {
            this.sceneReadyPromiseResolve = resolve;
        });
        this.settings = new ViewerSettings();
        this.components = new Components();
        this.fragmentIdToModelIdxMap = {}; // Used to identify the model id for picking etc.
        this.idMap = {};
        this.hoveredElement = null;
        this.hoveredElement$ = new BehaviorSubject(this.hoveredElement);
        this.fragmentTypeMap = {};
        this.loadedFragments = {};
        this.loadedFragments$ = new BehaviorSubject(this.loadedFragments);
        this.currentTool = Tool.NONE;
        this.currentTool$ = new BehaviorSubject(this.currentTool);
        this.loadedModels = [];
        this.loadedElements$ = new BehaviorSubject([]);
        this.selection = [];
        this.selection$ = new BehaviorSubject(this.selection);
    }
    // Initialize a scene
    init(canvas, settings) {
        if (!canvas)
            return;
        this.container = canvas;
        if (settings != undefined)
            this.settings = settings;
        const components = new Components();
        components.scene = new SimpleScene(components);
        const renderer = new PostproductionRenderer(components, canvas);
        renderer.postproduction.outlineColor = 0x999999;
        components.renderer = renderer;
        const camera = new OrthoPerspectiveCamera(components);
        camera.updateAspect();
        this.camera = camera;
        components.camera = this.camera;
        components.raycaster = new SimpleRaycaster(components);
        renderer.postproduction.setup(this.camera.controls);
        renderer.postproduction.active = true;
        components.init();
        this.components = components;
        // Add default lights
        this.addLights();
        // Add components like clipper, grid etc.
        this.addClipperComponent();
        this.addDimensionsComponent();
        this.addGridComponent();
        // Prepare to receive fragments
        this.initFragments();
        // Add event listeners
        this.addEvents();
        // Resolve scene ready promise
        this.sceneReadyPromiseResolve();
    }
    getClipper() {
        return this.clipper == undefined ? this.addClipperComponent() : this.clipper;
    }
    getGrid() {
        return this.grid == undefined ? this.addGridComponent() : this.grid;
    }
    getDimensions() {
        return this.dimensions == undefined ? this.addDimensionsComponent() : this.dimensions;
    }
    getCamera() {
        return this.camera;
    }
    // Removes all loaded geometry
    async clearScene() {
        await this.sceneReady;
        if (this.container == undefined)
            return;
        const scene = this.components.scene.get();
        if (this.fragments != undefined) {
            this.fragments.highlighter.active = false;
        }
        scene.clear();
        this.components.dispose();
        // Remove all children of container (except menu)
        const children = this.container.children;
        for (let i = 0; i < children.length; i++) {
            const el = children[i];
            if (!el.classList.contains("menu-container")) {
                this.container.removeChild(el);
            }
        }
        this.init(this.container);
    }
    // Load a fragments model
    async loadModel(file) {
        const index = this.loadedModels.length;
        const fileName = file.name.split("/").pop()?.split(".")[0] ?? index.toString();
        await this.sceneReady;
        const fragments = this.fragments ?? await this.initFragments();
        // Get zip content
        const url = URL.createObjectURL(file);
        const zipContent = await getZipContent(url, index, fileName);
        // Update global maps
        this.idMap = { ...this.idMap, ...zipContent.elementIdMap };
        this.fragmentTypeMap = { ...this.fragmentTypeMap, ...zipContent.fragmentTypeMap };
        this.fragmentIdToModelIdxMap = { ...this.fragmentIdToModelIdxMap, ...zipContent.fragmentIdToModelIdxMap };
        // Update loaded elements observable
        this.loadedElements$.next(Object.values(this.idMap));
        // Load the bounding boxes for this model
        fragments.memoryCuller.loadBoxes(url, zipContent.opaqueBoxes, zipContent.transparentBoxes, zipContent.fragmentMap);
        fragments.memoryCuller.needsUpdate = true;
        this.loadedModels.push({ index, fileName, folderPrefix: zipContent.folderPrefix, expressGlobalMap: zipContent.expressGlobalMap });
    }
    getHoveredElement() {
        return this.hoveredElement$.asObservable();
    }
    getSelection() {
        return this.selection$.asObservable();
    }
    getLoadedFragments() {
        return this.loadedFragments$.asObservable().pipe(debounceTime(this.settings?.refreshInterval ?? 500));
    }
    // Returns all the elements that have been loaded in ICDD zips.
    // Beware that some of them might not yet be discovered and rendered
    getLoadedElements() {
        return this.loadedElements$.asObservable();
    }
    getCurrentTool() {
        return this.currentTool$.asObservable();
    }
    setCurrentTool(tool) {
        this.currentTool = tool;
        this.currentTool$.next(tool);
    }
    updatePostProduction() {
        if (this.fragments == undefined)
            return;
        this.fragments.culler.needsUpdate = true;
        const renderer = this.components.renderer;
        renderer.postproduction.update();
    }
    // Load elements before they are seen by the culler
    async forceLoadElements(globalIds) {
        const map = await buildFragmentModelMap(globalIds, this.idMap);
        for (let zipURL of Object.keys(map)) {
            this.settings?.verbose && console.info(`Force loading ${map[zipURL].length} fragments`);
            const fragmentIds = map[zipURL];
            await this.loadFragments(zipURL, fragmentIds);
        }
    }
    async forceLoadElementsOfType(type) {
        const globalIds = await getAllOfType([type], this.idMap);
        return this.forceLoadElements(globalIds);
    }
    addLights() {
        const scene = this.components.scene.get();
        const directionalLight = new DirectionalLight();
        directionalLight.position.set(5, 10, 3);
        directionalLight.intensity = 0.5;
        scene.add(directionalLight);
        const ambientLight = new AmbientLight();
        ambientLight.intensity = 0.5;
        scene.add(ambientLight);
    }
    addClipperComponent() {
        const clipper = new EdgesClipper(this.components, EdgesPlane);
        clipper.enabled = true;
        this.components.tools.add(clipper);
        this.clipper = clipper;
        return clipper;
    }
    addDimensionsComponent() {
        const dimensions = new SimpleDimensions(this.components);
        this.components.tools.add(dimensions);
        dimensions.enabled = false;
        dimensions.snapDistance = 1;
        this.components.tools.add(dimensions);
        this.dimensions = dimensions;
        return dimensions;
    }
    async addEvents() {
        if (this.container == undefined)
            return;
        // Single click events
        this.container.onclick = async () => {
            // Stop if in tool mode
            if (this.currentTool != Tool.NONE)
                return;
            // If selection mode and hovering something pickable
            if (this.settings?.selectionMode != SelectionMode.DISABLED) {
                // Element click
                if (this.hoveredElement) {
                    const index = this.selection.map(item => item.expressID).indexOf(this.hoveredElement.expressID);
                    const removePrevious = this.settings?.selectionMode == SelectionMode.SINGLE;
                    // Already selected? Remove from selection
                    if (index != -1) {
                        removePrevious ? this.selection = [] : this.selection.splice(index, 1);
                        const map = await buildFragmentMapFromPicked(this.selection);
                        this.fragments?.highlighter.highlightByID('selection', map, true);
                    }
                    // Add to selection
                    else {
                        this.fragments?.highlighter.highlight('selection', removePrevious);
                        removePrevious ? this.selection = [this.hoveredElement] : this.selection.push(this.hoveredElement);
                    }
                    // Canvas click. Clear selection
                }
                else {
                    this.selection = [];
                    this.fragments?.highlighter.clear('selection');
                }
                this.selection$.next(this.selection);
            }
        };
        // Double click events
        this.container.ondblclick = () => {
            switch (this.currentTool) {
                // Add dimensions
                case Tool.DIMENSIONS:
                    this.dimensions?.create();
                    break;
                // Add clipping plane
                case Tool.CLIPPER:
                    this.clipper?.create();
                    break;
            }
        };
        // Mouse hover
        window.addEventListener('mousemove', async () => {
            // Do nothing if a tool is enabled enabled
            if (this.currentTool != Tool.NONE) {
                this.fragments?.highlighter.clear('highlight');
                return;
            }
            // See what is hovered
            const meshes = this.fragments?.fragmentMeshes;
            const result = this.components.raycaster.castRay(meshes);
            if (!result) {
                this.hoveredElement = null;
                this.fragments?.highlighter.clear('highlight');
            }
            else {
                const pickedObject = await this.getPickedObject(result);
                if (pickedObject.elementData?.unpickable) {
                    this.hoveredElement = null;
                    this.fragments?.highlighter.clear('highlight');
                }
                else {
                    this.hoveredElement = pickedObject;
                    // Append highlight color
                    if (this.settings?.highlightElementOnHover) {
                        this.fragments?.highlighter.highlight('highlight');
                    }
                }
            }
            this.hoveredElement$.next(this.hoveredElement);
        });
        // Window resize
        fromEvent(window, 'resize').pipe(debounceTime(500)).subscribe(() => {
            // Update aspect ratio
            if (this.camera != undefined)
                this.camera.updateAspect();
        });
    }
    addGridComponent() {
        const renderer = this.components.renderer;
        const grid = new SimpleGrid(this.components);
        grid.enabled = false;
        grid.get().visible = false;
        this.components.tools.add(grid);
        renderer.postproduction.excludedItems.add(grid.get());
        this.grid = grid;
        return grid;
    }
    async initFragments(container = this.container) {
        const renderer = this.components.renderer;
        const fragments = new Fragments(this.components);
        // Set up fragment culler update
        container.addEventListener('mouseup', () => fragments.culler.needsUpdate = true);
        container.addEventListener('wheel', () => fragments.culler.needsUpdate = true);
        // Set up highlighting
        fragments.highlighter.add('highlight', [new MeshBasicMaterial({ color: this.settings?.highlightColor })]);
        fragments.highlighter.add('selection', [new MeshBasicMaterial({ color: this.settings?.selectionColor })]);
        // Bind postproduction update with fragment culler update
        fragments.culler.viewUpdated.on(() => setTimeout(() => renderer.postproduction.update(), 300));
        this.fragments = fragments;
        const mainCanvas = renderer.get().domElement;
        mainCanvas.addEventListener('mousedown', () => {
            fragments.memoryCuller.needsUpdate = true;
        });
        // Load fragments only when the tiles are found in the screen
        fragments.memoryCuller.fragmentsDiscovered.on(async (discoveredFragments) => {
            for (const url in discoveredFragments) {
                const fragIDs = discoveredFragments[url];
                this.settings?.verbose && console.info(`Loading ${fragIDs.length} discovered fragments`);
                await this.loadFragments(url, fragIDs);
            }
        });
        return fragments;
    }
    // Used by the memory culler to continiously load fragments as they are discovered
    // Can also be used to force load important fragments. If the purpose is to show slabs
    // for example, it is possible to load those
    async loadFragments(zipURL, fragIDs) {
        if (this.fragments == undefined)
            return;
        const { entries } = await unzip(zipURL);
        for (const fragID of fragIDs) {
            if (this.loadedFragments[fragID] != undefined) {
                this.settings?.verbose && console.info(`Skipping fragment ${fragID}. Already loaded`);
                continue; // Already loaded
            }
            if (this.settings?.hideSpaces && this.fragmentTypeMap[fragID] == "IFCSPACE") {
                this.settings?.verbose && console.info(`Fragment ${fragID} is a space. Skipping`);
                continue;
            }
            const modelIndex = this.fragmentIdToModelIdxMap[fragID];
            const modelData = this.loadedModels[modelIndex];
            const geometry = await entries[`${modelData.folderPrefix}${fragID}.glb`].blob();
            const geometryURL = URL.createObjectURL(geometry);
            const dataBlob = await entries[`${modelData.folderPrefix}${fragID}.json`].blob();
            const dataURL = URL.createObjectURL(dataBlob);
            const fragment = await this.fragments.load(geometryURL, dataURL);
            this.loadedFragments[fragment.id] = fragment;
            this.loadedFragments$.next(this.loadedFragments);
            this.fragments.culler.needsUpdate = true;
            this.fragments.highlighter.update();
            this.fragments.highlighter.active = true;
        }
    }
    async getPickedObject(result) {
        let obj = { distance: result.distance, intersectionPoint: result.point };
        if (this.fragments == undefined)
            return obj;
        const fragmentMesh = result.object;
        const fragment = Object.values(this.fragments.fragments).find(frag => frag.mesh.uuid == fragmentMesh.uuid);
        if (fragment == undefined)
            return obj;
        const blockID = fragment.getVertexBlockID(result.object.geometry, result.face.a);
        const expressID = fragment.getItemID(result.instanceId, blockID);
        if (expressID != undefined)
            obj.expressID = parseInt(expressID);
        // Get element data only if it's not the same object as last time
        if (this.lastPicked != undefined && this.lastPicked.expressID == obj.expressID)
            return this.lastPicked;
        obj.elementData = await getElementData(expressID, this.idMap);
        this.lastPicked = obj;
        return obj;
    }
}
BimViewerService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: BimViewerService, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
BimViewerService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: BimViewerService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: BimViewerService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return []; } });

class MenuService {
    constructor(bimViewerService) {
        this.bimViewerService = bimViewerService;
        this.settings = new MenuSettings();
        this.settings$ = new BehaviorSubject(new MenuSettings());
    }
    getSettings() {
        return this.settings$.asObservable();
    }
    handleItemClick(id) {
        switch (id) {
            case 'clipper':
                this.settings.toggleClipperMode();
                if (!this.settings.clipperMode) {
                    this.disableClipper();
                }
                else {
                    this.enableClipper();
                }
                break;
            case 'dimensions':
                this.settings.toggleDimensionsMode();
                if (!this.settings.dimensionsMode) {
                    this.disableDimensions();
                }
                else {
                    this.enableDimensions();
                }
                break;
            case 'info':
                this.settings.toggleShowInfo();
                break;
            case 'grid':
                this.settings.toggleShowGrid();
                this.settings.showGrid ? this.showGrid() : this.hideGrid();
                break;
            case 'wipe':
                this.resetAllComponents();
                break;
            case 'orbit':
                this.orbitMode();
                this.settings.setNavMode('Orbit');
                break;
            case 'first-person':
                this.firstPersonMode();
                this.settings.setNavMode('FirstPerson');
                break;
            case 'plan':
                this.planMode();
                this.settings.setNavMode('Plan');
                break;
            case 'zoom-ext':
                this.zoomExtents();
                break;
        }
        this.updateSettings(this.settings);
    }
    resetAllComponents() {
        console.log("Reset all components");
        this.bimViewerService.setCurrentTool(Tool.NONE);
        // this.disableDimensions();
        this.removeDimensions();
        // this.disableClipper();
        this.removeClipper();
        this.settings.toggleDimensionsMode(false);
        this.settings.toggleClipperMode(false);
        this.settings.infoText = "";
    }
    // Clipper
    enableClipper() {
        this.bimViewerService.setCurrentTool(Tool.CLIPPER);
        document.body.style.cursor = 'default';
        this.settings.infoText = "Clipper mode is enabled. Double-click an element to create a section.";
    }
    disableClipper() {
        this.bimViewerService.setCurrentTool(Tool.NONE);
        this.settings.infoText = "";
    }
    removeClipper() {
        const clipper = this.bimViewerService.getClipper();
        clipper.deleteAll();
    }
    // Dimensions
    enableDimensions() {
        this.bimViewerService.setCurrentTool(Tool.DIMENSIONS);
        this.bimViewerService.getDimensions().enabled = true;
        document.body.style.cursor = 'default';
        this.settings.infoText = "Measurement mode is enabled. Double-click somewhere to start measuring.";
    }
    disableDimensions() {
        this.bimViewerService.setCurrentTool(Tool.NONE);
        this.bimViewerService.getDimensions().enabled = false;
        this.settings.infoText = "";
    }
    removeDimensions() {
        const dimensions = this.bimViewerService.getDimensions();
        dimensions.deleteAll();
    }
    // Camera
    zoomExtents() {
        const camera = this.bimViewerService.getCamera();
        camera.fitModelToFrame();
        this.bimViewerService.updatePostProduction();
    }
    firstPersonMode() {
        const camera = this.bimViewerService.getCamera();
        camera.setNavigationMode("FirstPerson");
    }
    orbitMode() {
        const camera = this.bimViewerService.getCamera();
        camera.setNavigationMode("Orbit");
    }
    planMode() {
        const camera = this.bimViewerService.getCamera();
        camera.setNavigationMode("Plan");
    }
    updateSettings(settings) {
        this.settings$.next(settings);
    }
    // Grid
    showGrid() {
        const grid = this.bimViewerService.getGrid();
        grid.enabled = true;
        grid.get().visible = true;
    }
    // NB! Not working
    hideGrid() {
        const grid = this.bimViewerService.getGrid();
        grid.enabled = false;
        grid.get().visible = false;
    }
}
MenuService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: MenuService, deps: [{ token: BimViewerService }], target: i0.ɵɵFactoryTarget.Injectable });
MenuService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: MenuService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: MenuService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: BimViewerService }]; } });

class IconsService {
    constructor(matIconRegistry, domSanitizer) {
        this.matIconRegistry = matIconRegistry;
        this.domSanitizer = domSanitizer;
    }
    loadIcons() {
        this.matIconRegistry.addSvgIcon(`clipper`, this.setPath('assets/icons/clipper.svg'));
        this.matIconRegistry.addSvgIcon(`clipper_toggled`, this.setPath('assets/icons/clipper_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`measure`, this.setPath('assets/icons/measure.svg'));
        this.matIconRegistry.addSvgIcon(`measure_toggled`, this.setPath('assets/icons/measure_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`zoom_ext`, this.setPath('assets/icons/zoom-ext.svg'));
        this.matIconRegistry.addSvgIcon(`info`, this.setPath('assets/icons/info.svg'));
        this.matIconRegistry.addSvgIcon(`info_toggled`, this.setPath('assets/icons/info_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`wipe`, this.setPath('assets/icons/wipe.svg'));
        this.matIconRegistry.addSvgIcon(`wipe_toggled`, this.setPath('assets/icons/wipe_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`hand`, this.setPath('assets/icons/hand.svg'));
        this.matIconRegistry.addSvgIcon(`hand_toggled`, this.setPath('assets/icons/hand_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`orbit`, this.setPath('assets/icons/orbit.svg'));
        this.matIconRegistry.addSvgIcon(`orbit_toggled`, this.setPath('assets/icons/orbit_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`first_person`, this.setPath('assets/icons/first_person.svg'));
        this.matIconRegistry.addSvgIcon(`first_person_toggled`, this.setPath('assets/icons/first_person_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`grid`, this.setPath('assets/icons/grid.svg'));
        this.matIconRegistry.addSvgIcon(`grid_toggled`, this.setPath('assets/icons/grid_toggled.svg'));
        this.matIconRegistry.addSvgIcon(`visibility`, this.setPath('assets/icons/visibility.svg'));
    }
    setPath(url) {
        return this.domSanitizer.bypassSecurityTrustResourceUrl(url);
    }
}
IconsService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: IconsService, deps: [{ token: i1.MatIconRegistry }, { token: i2.DomSanitizer }], target: i0.ɵɵFactoryTarget.Injectable });
IconsService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: IconsService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: IconsService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: i1.MatIconRegistry }, { type: i2.DomSanitizer }]; } });

class VisibilityService {
    constructor(bimViewerService) {
        this.bimViewerService = bimViewerService;
        this.loadedFragments = {};
        this.originalMaterials = {}; // Map with original materials of all fragments
        this.dimmedMaterials = {}; // Map with dimmed materials of all fragments
        this.groups$ = new BehaviorSubject(this.groupData);
        this.hiddenElements = [];
        // Build material maps every time we receive new fragments from the viewer
        this.bimViewerService.getLoadedFragments().subscribe((loadedFragments) => {
            this.loadedFragments = loadedFragments;
            this.buildMaterialMaps();
            this.reApplyVisibility();
            this.buildGroups();
        });
    }
    /**
     * Dimming
     */
    async dimAll() {
        this.allDimmed = true;
        markAllUnpickable(this.bimViewerService.idMap);
        await this.reApplyVisibility();
    }
    async undimAll() {
        this.allDimmed = false;
        markAllPickable(this.bimViewerService.idMap);
        await this.reApplyVisibility();
    }
    /**
     * Hiding
     */
    async hideAll() {
        this.allHidden = true;
        await this.reApplyVisibility();
    }
    async unhideAll() {
        this.allHidden = false;
        await this.reApplyVisibility();
    }
    async hideElements(globalIds) {
        this.allHidden = undefined;
        this.hiddenElements = globalIds;
        await this.reApplyVisibility();
    }
    async isolateElements(globalIds) {
        this.allHidden = undefined;
        this.hiddenElements = await getAllRemainingGlobalIds(globalIds, this.bimViewerService.idMap);
        await this.bimViewerService.forceLoadElements(globalIds); // Force load them so we are sure they all show up
        await this.reApplyVisibility();
    }
    // Hides all elements that belong to any of the types specified in the array (eg. ["IFCSPACE"])
    async hideAllOfType(types) {
        const globalIds = await getAllOfType(types, this.bimViewerService.idMap);
        await this.hideElements(globalIds);
    }
    // Isolates all elements (hides remaining) that belong to any of the types specified in the array (eg. ["IFCSPACE"])
    async isolateAllOfType(types) {
        const [visible, hidden] = await Promise.all([
            getAllOfType(types, this.bimViewerService.idMap),
            getAllNotOfType(types, this.bimViewerService.idMap)
        ]);
        await this.bimViewerService.forceLoadElements(visible); // Force load them so we are sure they all show up
        await this.hideElements(hidden);
    }
    getVisibilityGroups() {
        return this.groups$.asObservable().pipe(debounceTime(this.bimViewerService.settings?.refreshInterval ?? 500) // Update at max every x ms
        );
    }
    // NB! This should probably be reapplied as new fragments appear
    async setGroupVisibility(category, name, visible) {
        return new Promise(async (resolve) => {
            const fragments = this.bimViewerService.fragments;
            if (fragments == undefined)
                resolve();
            if (fragments == undefined)
                return;
            if (this.groupData == undefined)
                return;
            const models = fragments.groups.get({ [category]: name });
            for (const guid in models) {
                const ids = models[guid];
                const frag = fragments.fragments[guid];
                if (frag == undefined)
                    continue;
                frag.setVisibility(ids, visible);
                const culled = fragments.culler.meshes.get(frag.id);
                if (culled)
                    culled.count = frag.mesh.count;
            }
            this.groupData.visibility[category][name] = visible;
            this.groups$.next(this.groupData);
            this.bimViewerService.updatePostProduction();
        });
    }
    // To be fired each time new fragments are discovered while navigating in the scene
    async reApplyVisibility() {
        return new Promise(async (resolve) => {
            const fragments = this.bimViewerService.fragments;
            if (fragments == undefined)
                resolve();
            // Dimming
            if (this.allDimmed === true) {
                Object.values(this.loadedFragments).forEach((fragment) => {
                    fragment.mesh.material = this.dimmedMaterials[fragment.id];
                });
            }
            else if (this.allDimmed === false) {
                Object.values(this.loadedFragments).forEach((fragment) => {
                    fragment.mesh.material = this.originalMaterials[fragment.id];
                });
            }
            // Hiding
            if (this.allHidden === true) {
                Object.values(this.loadedFragments).forEach((fragment) => {
                    fragment.setVisibility(fragment.items, false);
                });
            }
            else if (this.allHidden === false) {
                Object.values(this.loadedFragments).forEach((fragment) => {
                    fragment.resetVisibility();
                });
            }
            else if (this.hiddenElements.length) {
                const map = await buildFragmentMap(this.hiddenElements, this.bimViewerService.idMap);
                Object.keys(map).forEach(fragmentId => {
                    const fragment = this.loadedFragments[fragmentId];
                    if (fragment != undefined)
                        fragment.setVisibility(map[fragmentId], false);
                });
            }
            resolve();
        });
    }
    // Tested and it is really speedy (7 ms on big model) + faster second time (>1ms on big model)
    buildMaterialMaps() {
        if (this.loadedFragments == undefined)
            return;
        Object.values(this.loadedFragments).forEach((fragment) => {
            // Save undimmed material of fragment if not already saved
            if (this.originalMaterials[fragment.id] == undefined) {
                this.originalMaterials[fragment.id] = Array.isArray(fragment.mesh.material) ? fragment.mesh.material.map((item) => item.clone()) : fragment.mesh.material.clone();
                this.dimmedMaterials[fragment.id] = Array.isArray(fragment.mesh.material) ? fragment.mesh.material.map((item) => item.clone()) : fragment.mesh.material.clone();
                if (Array.isArray(this.dimmedMaterials[fragment.id])) {
                    this.dimmedMaterials[fragment.id].forEach((mat) => {
                        mat.transparent = true;
                        mat.opacity = 0.1;
                    });
                }
                else {
                    this.dimmedMaterials[fragment.id].transparent = true;
                    this.dimmedMaterials[fragment.id].opacity = 0.1;
                }
            }
        });
    }
    // Each fragment should be assigned with its model, category and potentially also storey
    async buildGroups() {
        const fragments = this.bimViewerService.fragments;
        if (fragments == undefined)
            return;
        // Group items by category and by floor
        const categoryID = "category";
        const floorID = "floor";
        const modelID = "model";
        const groups = { [categoryID]: {}, [floorID]: {}, [modelID]: {} };
        const visibilities = this.groupData?.visibility ?? { [categoryID]: {}, [floorID]: {}, [modelID]: {} };
        Object.values(this.bimViewerService.idMap).forEach((item) => {
            // Specify category group
            if (item.type != undefined) {
                if (!groups[categoryID][item.type]) {
                    groups[categoryID][item.type] = [];
                }
                if (visibilities[categoryID][item.type] == undefined) {
                    visibilities[categoryID][item.type] = true;
                }
                groups[categoryID][item.type].push(item.expressId);
            }
            // Specify level group
            if (item.levelId != undefined) {
                const lvlName = item.levelData?.name ?? item.levelId;
                if (!groups[floorID][lvlName]) {
                    groups[floorID][lvlName] = [];
                }
                if (visibilities[floorID][lvlName] == undefined) {
                    visibilities[floorID][lvlName] = true;
                }
                groups[floorID][lvlName].push(item.expressId);
            }
            // Specify model group
            if (!groups[modelID][item.modelName]) {
                groups[modelID][item.modelName] = [];
            }
            if (visibilities[modelID][item.modelName] == undefined) {
                visibilities[modelID][item.modelName] = true;
            }
            groups[modelID][item.modelName].push(item.expressId);
            fragments.groups.add(item.fragmentId, groups);
        });
        this.groupData = { groupSystems: fragments.groups.groupSystems, visibility: visibilities };
        this.groups$.next(this.groupData);
    }
}
VisibilityService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: VisibilityService, deps: [{ token: BimViewerService }], target: i0.ɵɵFactoryTarget.Injectable });
VisibilityService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: VisibilityService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: VisibilityService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: BimViewerService }]; } });

class MenuComponent {
    constructor(menuService, iconsService, visibilityService) {
        this.menuService = menuService;
        this.iconsService = iconsService;
        this.visibilityService = visibilityService;
        this.gotIcons = false;
        this.settings = new MenuSettings();
        this.visibilityMenu = new VisibilityMenu();
        this.onItemClick = new EventEmitter();
    }
    ngOnInit() {
        this.iconsService.loadIcons();
        this.gotIcons = true;
        // Monitor settings
        this.menuService.getSettings().subscribe(settings => this.settings = settings);
        // Monitor groups
        this.visibilityService.getVisibilityGroups().subscribe(groupData => this.groupData = groupData);
    }
    handleItemClick(id) {
        this.menuService.handleItemClick(id);
        this.onItemClick.emit({ id, settings: this.settings });
    }
    closeAllMenus() {
        this.visibilityMenu.visible = false;
    }
    showVisibilityMenu(ev) {
        this.buildVisibilityMenu(ev.pageX, ev.pageY);
    }
    buildVisibilityMenu(x, y) {
        if (this.groupData == undefined)
            return;
        const groupSystems = this.groupData.groupSystems;
        this.visibilityMenu = {
            left: `${x}px`,
            top: `${y}px`,
            visible: !this.visibilityMenu.visible,
            categories: Object.keys(groupSystems).map(catName => {
                const instanceNames = Object.keys(groupSystems[catName]);
                const instanceGroups = instanceNames.map(instName => {
                    const visible = this.groupData?.visibility?.[catName]?.[instName] ?? true;
                    return { visible, name: instName, instanceCount: groupSystems[catName][instName].length };
                });
                return { name: catName, unfolded: false, instanceGroups };
            }) ?? [],
        };
    }
    toggleUnfold(category) {
        const currentSetting = category.unfolded;
        this.visibilityMenu.categories.forEach(cat => cat.unfolded = false);
        category.unfolded = !currentSetting;
    }
    setGroupVisibility(category, name, visible) {
        this.visibilityService.setGroupVisibility(category, name, visible);
    }
    onKeydownHandler(event) {
        if (event.key == "Escape") {
            this.menuService.handleItemClick("wipe");
        }
    }
}
MenuComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: MenuComponent, deps: [{ token: MenuService }, { token: IconsService }, { token: VisibilityService }], target: i0.ɵɵFactoryTarget.Component });
MenuComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "15.1.0", type: MenuComponent, selector: "app-menu", outputs: { onItemClick: "onItemClick" }, host: { listeners: { "document:keydown": "onKeydownHandler($event)" } }, ngImport: i0, template: "<div class=\"container\" *ngIf=\"gotIcons\">\r\n    <div class=\"info-text\" *ngIf=\"settings.showInfo\">{{settings.infoText}}</div>\r\n    <div class=\"menu-container\">\r\n\r\n        <!-- Tool group -->\r\n        <div class=\"group-container\">\r\n\r\n            <!-- Clipper mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('clipper')\"\r\n                matTooltip=\"Toggle clipper mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.clipperMode ? 'clipper_toggled' : 'clipper'\">\r\n            </mat-icon>\r\n\r\n            <!-- Dimensions mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('dimensions')\"\r\n                matTooltip=\"Toggle measurement mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.dimensionsMode ? 'measure_toggled' : 'measure'\">\r\n            </mat-icon>\r\n\r\n            <!-- Wipe all -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('wipe')\"\r\n                matTooltip=\"Wipe sections and measurements\" matTooltipPosition=\"above\"\r\n                svgIcon=\"wipe\">\r\n            </mat-icon>\r\n\r\n        </div>\r\n\r\n        <!-- Navigation group -->\r\n        <div class=\"group-container\">\r\n\r\n            <!-- Orbit mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('orbit')\"\r\n                matTooltip=\"Orbit navigation mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.orbitMode ? 'orbit_toggled' : 'orbit'\">\r\n            </mat-icon>\r\n\r\n            <!-- Plan mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('plan')\"\r\n                matTooltip=\"Plan navigation mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.planMode ? 'hand_toggled' : 'hand'\">\r\n            </mat-icon>\r\n\r\n            <!-- First person mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('first-person')\"\r\n                matTooltip=\"First person navigation mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.firstPersonMode ? 'first_person_toggled' : 'first_person'\">\r\n            </mat-icon>\r\n\r\n            <!-- Zoom extents -->\r\n            <mat-icon class=\"btn\" matTooltip=\"Zoom extents\" matTooltipPosition=\"above\"\r\n                (click)=\"handleItemClick('zoom-ext')\" svgIcon=\"zoom_ext\"></mat-icon>\r\n\r\n        </div>\r\n\r\n        <!-- Tool Visibility -->\r\n        <div class=\"group-container\">\r\n\r\n            <!-- Show info -->\r\n            <mat-icon class=\"btn\" matTooltip=\"Display info texts?\" matTooltipPosition=\"above\"\r\n                (click)=\"handleItemClick('info')\" [svgIcon]=\"settings.showInfo ? 'info_toggled' : 'info'\"></mat-icon>\r\n            \r\n            <!-- Show grid -->\r\n            <mat-icon class=\"btn\" matTooltip=\"Display grid?\" matTooltipPosition=\"above\"\r\n                (click)=\"handleItemClick('grid')\" [svgIcon]=\"settings.showGrid ? 'grid_toggled' : 'grid'\"></mat-icon>\r\n\r\n            <!-- Visibility -->\r\n            <mat-icon class=\"btn\" svgIcon=\"visibility\" (click)=\"showVisibilityMenu($event)\"></mat-icon>\r\n\r\n            <!-- Visibility menu -->\r\n            <div *ngIf=\"visibilityMenu.visible\" \r\n                [style.left]=\"visibilityMenu.left\" \r\n                [style.top]=\"visibilityMenu.top\"\r\n                class=\"context-menu\">\r\n                <div class=\"menu\">\r\n                    <div *ngFor=\"let category of visibilityMenu.categories\">\r\n                        <div class=\"row btn\" (click)=\"toggleUnfold(category)\">\r\n                            {{category.name}} \r\n                            <mat-icon *ngIf=\"!category.unfolded\">expand_more</mat-icon>\r\n                            <mat-icon *ngIf=\"category.unfolded\">expand_less</mat-icon>\r\n                        </div>\r\n                        <ng-container *ngIf=\"category.unfolded\">\r\n                            <div class=\"row inst\" *ngFor=\"let instance of category.instanceGroups\">\r\n                                {{instance.name}}\r\n                                <input type=\"checkbox\" [(ngModel)]=\"instance.visible\" (change)=\"setGroupVisibility(category.name, instance.name, instance.visible)\">\r\n                            </div>\r\n                        </ng-container>\r\n                    </div>\r\n                </div>\r\n            </div>\r\n\r\n        </div>\r\n\r\n    </div>\r\n</div>", styles: [".container{display:flex;flex-direction:column;gap:5px}.menu-container{display:flex;flex-direction:row;justify-content:center;gap:5px;width:100%}.group-container{background-color:#353535cc;display:flex;flex-direction:row;gap:10px;padding:10px;border-radius:5px}.btn{color:#fff}.btn:hover{cursor:pointer;opacity:.8}.toggled{color:#2c2c64}.info-text{font-size:.8em}.context-menu{position:fixed;display:inline-block;transform:translate(-100%);z-index:100}.menu{display:flex;flex-direction:column;color:#fff;background-color:#353535cc;gap:10px;padding:10px;border-radius:5px}.row{display:flex;flex-direction:row;gap:5px;justify-content:space-between}.inst{font-size:.8em}\n"], dependencies: [{ kind: "directive", type: i4.NgForOf, selector: "[ngFor][ngForOf]", inputs: ["ngForOf", "ngForTrackBy", "ngForTemplate"] }, { kind: "directive", type: i4.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }, { kind: "directive", type: i5.CheckboxControlValueAccessor, selector: "input[type=checkbox][formControlName],input[type=checkbox][formControl],input[type=checkbox][ngModel]" }, { kind: "directive", type: i5.NgControlStatus, selector: "[formControlName],[ngModel],[formControl]" }, { kind: "directive", type: i5.NgModel, selector: "[ngModel]:not([formControlName]):not([formControl])", inputs: ["name", "disabled", "ngModel", "ngModelOptions"], outputs: ["ngModelChange"], exportAs: ["ngModel"] }, { kind: "component", type: i1.MatIcon, selector: "mat-icon", inputs: ["color", "inline", "svgIcon", "fontSet", "fontIcon"], exportAs: ["matIcon"] }, { kind: "directive", type: i7.MatLegacyTooltip, selector: "[matTooltip]", exportAs: ["matTooltip"] }] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: MenuComponent, decorators: [{
            type: Component,
            args: [{ selector: 'app-menu', template: "<div class=\"container\" *ngIf=\"gotIcons\">\r\n    <div class=\"info-text\" *ngIf=\"settings.showInfo\">{{settings.infoText}}</div>\r\n    <div class=\"menu-container\">\r\n\r\n        <!-- Tool group -->\r\n        <div class=\"group-container\">\r\n\r\n            <!-- Clipper mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('clipper')\"\r\n                matTooltip=\"Toggle clipper mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.clipperMode ? 'clipper_toggled' : 'clipper'\">\r\n            </mat-icon>\r\n\r\n            <!-- Dimensions mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('dimensions')\"\r\n                matTooltip=\"Toggle measurement mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.dimensionsMode ? 'measure_toggled' : 'measure'\">\r\n            </mat-icon>\r\n\r\n            <!-- Wipe all -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('wipe')\"\r\n                matTooltip=\"Wipe sections and measurements\" matTooltipPosition=\"above\"\r\n                svgIcon=\"wipe\">\r\n            </mat-icon>\r\n\r\n        </div>\r\n\r\n        <!-- Navigation group -->\r\n        <div class=\"group-container\">\r\n\r\n            <!-- Orbit mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('orbit')\"\r\n                matTooltip=\"Orbit navigation mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.orbitMode ? 'orbit_toggled' : 'orbit'\">\r\n            </mat-icon>\r\n\r\n            <!-- Plan mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('plan')\"\r\n                matTooltip=\"Plan navigation mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.planMode ? 'hand_toggled' : 'hand'\">\r\n            </mat-icon>\r\n\r\n            <!-- First person mode -->\r\n            <mat-icon class=\"btn\" (click)=\"handleItemClick('first-person')\"\r\n                matTooltip=\"First person navigation mode\" matTooltipPosition=\"above\"\r\n                [svgIcon]=\"settings.firstPersonMode ? 'first_person_toggled' : 'first_person'\">\r\n            </mat-icon>\r\n\r\n            <!-- Zoom extents -->\r\n            <mat-icon class=\"btn\" matTooltip=\"Zoom extents\" matTooltipPosition=\"above\"\r\n                (click)=\"handleItemClick('zoom-ext')\" svgIcon=\"zoom_ext\"></mat-icon>\r\n\r\n        </div>\r\n\r\n        <!-- Tool Visibility -->\r\n        <div class=\"group-container\">\r\n\r\n            <!-- Show info -->\r\n            <mat-icon class=\"btn\" matTooltip=\"Display info texts?\" matTooltipPosition=\"above\"\r\n                (click)=\"handleItemClick('info')\" [svgIcon]=\"settings.showInfo ? 'info_toggled' : 'info'\"></mat-icon>\r\n            \r\n            <!-- Show grid -->\r\n            <mat-icon class=\"btn\" matTooltip=\"Display grid?\" matTooltipPosition=\"above\"\r\n                (click)=\"handleItemClick('grid')\" [svgIcon]=\"settings.showGrid ? 'grid_toggled' : 'grid'\"></mat-icon>\r\n\r\n            <!-- Visibility -->\r\n            <mat-icon class=\"btn\" svgIcon=\"visibility\" (click)=\"showVisibilityMenu($event)\"></mat-icon>\r\n\r\n            <!-- Visibility menu -->\r\n            <div *ngIf=\"visibilityMenu.visible\" \r\n                [style.left]=\"visibilityMenu.left\" \r\n                [style.top]=\"visibilityMenu.top\"\r\n                class=\"context-menu\">\r\n                <div class=\"menu\">\r\n                    <div *ngFor=\"let category of visibilityMenu.categories\">\r\n                        <div class=\"row btn\" (click)=\"toggleUnfold(category)\">\r\n                            {{category.name}} \r\n                            <mat-icon *ngIf=\"!category.unfolded\">expand_more</mat-icon>\r\n                            <mat-icon *ngIf=\"category.unfolded\">expand_less</mat-icon>\r\n                        </div>\r\n                        <ng-container *ngIf=\"category.unfolded\">\r\n                            <div class=\"row inst\" *ngFor=\"let instance of category.instanceGroups\">\r\n                                {{instance.name}}\r\n                                <input type=\"checkbox\" [(ngModel)]=\"instance.visible\" (change)=\"setGroupVisibility(category.name, instance.name, instance.visible)\">\r\n                            </div>\r\n                        </ng-container>\r\n                    </div>\r\n                </div>\r\n            </div>\r\n\r\n        </div>\r\n\r\n    </div>\r\n</div>", styles: [".container{display:flex;flex-direction:column;gap:5px}.menu-container{display:flex;flex-direction:row;justify-content:center;gap:5px;width:100%}.group-container{background-color:#353535cc;display:flex;flex-direction:row;gap:10px;padding:10px;border-radius:5px}.btn{color:#fff}.btn:hover{cursor:pointer;opacity:.8}.toggled{color:#2c2c64}.info-text{font-size:.8em}.context-menu{position:fixed;display:inline-block;transform:translate(-100%);z-index:100}.menu{display:flex;flex-direction:column;color:#fff;background-color:#353535cc;gap:10px;padding:10px;border-radius:5px}.row{display:flex;flex-direction:row;gap:5px;justify-content:space-between}.inst{font-size:.8em}\n"] }]
        }], ctorParameters: function () { return [{ type: MenuService }, { type: IconsService }, { type: VisibilityService }]; }, propDecorators: { onItemClick: [{
                type: Output
            }], onKeydownHandler: [{
                type: HostListener,
                args: ['document:keydown', ['$event']]
            }] } });

class BimViewerComponent {
    constructor(bimViewerService, menuService) {
        this.bimViewerService = bimViewerService;
        this.menuService = menuService;
        this.setClipperMode = true;
        this.settings = new ViewerSettings();
        this.hoveredElement = new EventEmitter();
        this.clickedElement = new EventEmitter();
        this.clickedCanvas = new EventEmitter();
        this.rightClickedElement = new EventEmitter();
        this.rightClickedCanvas = new EventEmitter();
        this.selectionChanged = new EventEmitter();
        this.orbiting = false;
        this.panning = false;
        this.currentlyHoveredElement = null;
    }
    set content(content) {
        if (content) {
            this.canvas = content.nativeElement;
            // initially setter gets called with undefined
            this.bimViewerService.init(content.nativeElement, this.settings);
        }
    }
    ngOnInit() {
        // Monitor hovered element
        this.bimViewerService.getHoveredElement().subscribe(el => {
            this.currentlyHoveredElement = el;
            // If hovering an element
            if (el && !this.menuSettings?.clipperMode && !this.menuSettings?.dimensionsMode) {
                if (this.settings.showPointerCursorOnElementHover)
                    document.body.style.cursor = 'pointer';
            }
            else {
                if (this.settings.showPointerCursorOnElementHover)
                    document.body.style.cursor = 'default';
            }
        });
        // Monitor menu state
        this.menuService.getSettings().subscribe(settings => this.menuSettings = settings);
        // Monitor selection
        this.bimViewerService.getSelection().subscribe(selection => {
            console.log(selection);
            this.selectionChanged.emit(selection);
        });
    }
    onLongPress(ev) {
        this.orbiting = true;
    }
    onLongPressRight(ev) {
        this.panning = true;
    }
    onShortPress(ev) {
        // Do nothing if in clipper or dimensions mode
        if (this.menuSettings?.clipperMode || this.menuSettings?.dimensionsMode)
            return;
        if (this.currentlyHoveredElement) {
            this.clickedElement.emit(this.currentlyHoveredElement);
        }
        else if (ev.target instanceof HTMLCanvasElement) {
            this.menuComponent.closeAllMenus();
            this.clickedCanvas.emit(ev);
        }
    }
    onShortPressRight(ev) {
        if (this.currentlyHoveredElement != undefined) {
            this.rightClickedElement.emit(this.currentlyHoveredElement);
        }
        else {
            this.rightClickedCanvas.emit(ev);
        }
    }
    handleMenuItemClick(ev) {
        // console.log(ev);
    }
    onMouseUp() {
        this.orbiting = false;
        this.panning = false;
    }
}
BimViewerComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: BimViewerComponent, deps: [{ token: BimViewerService }, { token: MenuService }], target: i0.ɵɵFactoryTarget.Component });
BimViewerComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "15.1.0", type: BimViewerComponent, selector: "lbd-fragment-viewer", inputs: { setClipperMode: "setClipperMode", settings: "settings" }, outputs: { hoveredElement: "hoveredElement", clickedElement: "clickedElement", clickedCanvas: "clickedCanvas", rightClickedElement: "rightClickedElement", rightClickedCanvas: "rightClickedCanvas", selectionChanged: "selectionChanged" }, host: { listeners: { "mouseup": "onMouseUp()" } }, viewQueries: [{ propertyName: "content", first: true, predicate: ["canvas"], descendants: true }, { propertyName: "menuComponent", first: true, predicate: MenuComponent, descendants: true }], ngImport: i0, template: "<div \r\n    appLongPress\r\n    appLongPressRight\r\n    appShortPress\r\n    appShortPressRight\r\n    [class.orbiting]=\"orbiting\"\r\n    [class.panning]=\"panning\"\r\n    #canvas \r\n    id=\"three-canvas\"\r\n    (longPress)=\"onLongPress($event)\"\r\n    (longPressRight)=\"onLongPressRight($event)\"\r\n    (shortPress)=\"onShortPress($event)\"\r\n    (shortPressRight)=\"onShortPressRight($event)\">\r\n\r\n    <div class=\"menu-container\">\r\n        <app-menu (onItemClick)=\"handleMenuItemClick($event)\"></app-menu>\r\n    </div>\r\n    \r\n</div>", styles: ["#three-canvas{position:relative;height:calc(100% - 10px);width:100%;overflow:hidden;z-index:1}.menu-container{position:absolute;display:flex;flex-direction:row;width:100%;justify-content:center;bottom:5px;z-index:100}::ng-deep .ifcjs-dimension-label{background-color:#000;font-family:sans-serif;color:#fff;padding:8px;border-radius:8px;pointer-events:all;transition:background-color .2s ease-in-out}::ng-deep .ifcjs-dimension-label:hover{background-color:gray}::ng-deep .ifcjs-dimension-preview{background-color:#fff;width:2rem;height:2rem;opacity:.3;padding:8px;border-radius:100%}\n"], dependencies: [{ kind: "component", type: MenuComponent, selector: "app-menu", outputs: ["onItemClick"] }] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: BimViewerComponent, decorators: [{
            type: Component,
            args: [{ selector: 'lbd-fragment-viewer', template: "<div \r\n    appLongPress\r\n    appLongPressRight\r\n    appShortPress\r\n    appShortPressRight\r\n    [class.orbiting]=\"orbiting\"\r\n    [class.panning]=\"panning\"\r\n    #canvas \r\n    id=\"three-canvas\"\r\n    (longPress)=\"onLongPress($event)\"\r\n    (longPressRight)=\"onLongPressRight($event)\"\r\n    (shortPress)=\"onShortPress($event)\"\r\n    (shortPressRight)=\"onShortPressRight($event)\">\r\n\r\n    <div class=\"menu-container\">\r\n        <app-menu (onItemClick)=\"handleMenuItemClick($event)\"></app-menu>\r\n    </div>\r\n    \r\n</div>", styles: ["#three-canvas{position:relative;height:calc(100% - 10px);width:100%;overflow:hidden;z-index:1}.menu-container{position:absolute;display:flex;flex-direction:row;width:100%;justify-content:center;bottom:5px;z-index:100}::ng-deep .ifcjs-dimension-label{background-color:#000;font-family:sans-serif;color:#fff;padding:8px;border-radius:8px;pointer-events:all;transition:background-color .2s ease-in-out}::ng-deep .ifcjs-dimension-label:hover{background-color:gray}::ng-deep .ifcjs-dimension-preview{background-color:#fff;width:2rem;height:2rem;opacity:.3;padding:8px;border-radius:100%}\n"] }]
        }], ctorParameters: function () { return [{ type: BimViewerService }, { type: MenuService }]; }, propDecorators: { setClipperMode: [{
                type: Input
            }], settings: [{
                type: Input
            }], hoveredElement: [{
                type: Output
            }], clickedElement: [{
                type: Output
            }], clickedCanvas: [{
                type: Output
            }], rightClickedElement: [{
                type: Output
            }], rightClickedCanvas: [{
                type: Output
            }], selectionChanged: [{
                type: Output
            }], content: [{
                type: ViewChild,
                args: ['canvas']
            }], menuComponent: [{
                type: ViewChild,
                args: [MenuComponent]
            }], onMouseUp: [{
                type: HostListener,
                args: ['mouseup']
            }] } });

class ColorService {
    constructor(bimViewerService) {
        this.bimViewerService = bimViewerService;
        this.loadedFragments = {};
        this.highlightMats = {};
        this.globalIds = {};
        this.selection = {};
        this.tempMatrix = new Matrix4();
        // Build material maps every time we receive new fragments from the viewer
        this.bimViewerService.getLoadedFragments().subscribe(async (loadedFragments) => {
            this.loadedFragments = loadedFragments;
            await this.reApplyColors();
        });
    }
    // Resets all colors
    async resetAllColors() {
        this.highlightMats = {};
        this.globalIds = {};
        this.selection = {};
        this.reApplyColors();
    }
    // Sets the color definition and applies the colors as the elements become visible in the scene
    async colorElements(colorDefinitions, resetPrevious = false) {
        let counter = Object.keys(this.globalIds).length;
        if (resetPrevious) {
            this.highlightMats = {};
            this.globalIds = {};
            this.selection = {};
            counter = 0;
        }
        // Build groups for all colors
        let globalIds = [];
        colorDefinitions.forEach(colorDef => {
            const id = `color${counter}`;
            this.addColorGroup(id, [new MeshBasicMaterial({ color: colorDef.color })]);
            this.globalIds[id] = colorDef.globalIds;
            globalIds.push(...colorDef.globalIds);
            markPickable(colorDef.globalIds, this.bimViewerService.idMap);
            counter++;
        });
        await this.bimViewerService.forceLoadElements(globalIds); // Force load them so we are sure they all show up
        await this.reApplyColors();
    }
    // Colors all elements that belong to a specific type (eg. "IFCSPACE")
    async colorAllOfType(type, color, resetPrevious = false) {
        const globalIds = await getAllOfType([type], this.bimViewerService.idMap);
        const colorDefinition = { globalIds, color };
        this.colorElements([colorDefinition], resetPrevious);
    }
    // To be fired each time new fragments are discovered while navigating in the scene
    async reApplyColors() {
        return new Promise(async (resolve) => {
            const fragments = this.bimViewerService.fragments;
            if (fragments == undefined)
                resolve();
            this.update();
            for (let id of Object.keys(this.globalIds)) {
                const map = await buildFragmentMap(this.globalIds[id], this.bimViewerService.idMap);
                Object.keys(map).forEach(fragmentId => {
                    const fragment = this.loadedFragments[fragmentId];
                    if (fragment != undefined) {
                        const mesh = fragment.mesh;
                        this.selection[id] = {};
                        this.selection[id][mesh.uuid] = map[fragmentId];
                        this.updateFragmentHighlight(id, mesh.uuid);
                    }
                });
            }
            resolve();
        });
    }
    addColorGroup(name, material) {
        if (this.highlightMats[name]) {
            throw new Error("A highlight with this name already exists.");
        }
        this.highlightMats[name] = material;
        this.selection[name] = {};
        this.update();
    }
    update() {
        const fragments = this.bimViewerService.fragments;
        if (fragments == undefined)
            return;
        for (const fragmentID in fragments.fragments) {
            const fragment = fragments.fragments[fragmentID];
            this.addHighlightToFragment(fragment);
        }
    }
    addHighlightToFragment(fragment) {
        for (const name in this.highlightMats) {
            if (!fragment.fragments[name]) {
                const material = this.highlightMats[name];
                fragment.addFragment(name, material);
                fragment.fragments[name].mesh.renderOrder = 0.5;
            }
        }
    }
    updateFragmentHighlight(name, fragmentID) {
        const fragments = this.bimViewerService.fragments;
        const components = this.bimViewerService.components;
        if (fragments == undefined)
            return;
        const ids = this.selection[name][fragmentID];
        const fragment = fragments.fragments[fragmentID];
        const selection = fragment.fragments[name];
        const scene = components.scene.get();
        const isBlockFragment = selection.blocks.count > 1;
        scene.add(selection.mesh);
        let i = 0;
        const blockIDs = [];
        for (const id of ids) {
            const { instanceID, blockID } = fragment.getInstanceAndBlockID(id);
            fragment.getInstance(instanceID, this.tempMatrix);
            try {
                selection.setInstance(i, { ids: [id], transform: this.tempMatrix });
            }
            catch (err) {
                // console.log(err);
            }
            if (isBlockFragment) {
                blockIDs.push(blockID);
            }
            i++;
        }
        if (isBlockFragment) {
            selection.blocks.add(blockIDs, true);
        }
        selection.mesh.count = i;
    }
}
ColorService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: ColorService, deps: [{ token: BimViewerService }], target: i0.ɵɵFactoryTarget.Injectable });
ColorService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: ColorService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: ColorService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: BimViewerService }]; } });

class LBDComponentsModule {
}
LBDComponentsModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: LBDComponentsModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
LBDComponentsModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "15.1.0", ngImport: i0, type: LBDComponentsModule, declarations: [BimViewerComponent,
        MenuComponent], imports: [CommonModule,
        FormsModule,
        HttpClientModule,
        MatIconModule,
        MatLegacyButtonModule,
        MatLegacyTooltipModule], exports: [BimViewerComponent] });
LBDComponentsModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: LBDComponentsModule, providers: [
        LBDComponentsService,
        SerializerService,
        BimViewerService,
        ColorService,
        VisibilityService
    ], imports: [CommonModule,
        FormsModule,
        HttpClientModule,
        MatIconModule,
        MatLegacyButtonModule,
        MatLegacyTooltipModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.1.0", ngImport: i0, type: LBDComponentsModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [
                        CommonModule,
                        FormsModule,
                        HttpClientModule,
                        MatIconModule,
                        MatLegacyButtonModule,
                        MatLegacyTooltipModule
                    ],
                    exports: [
                        BimViewerComponent
                    ],
                    declarations: [
                        BimViewerComponent,
                        MenuComponent
                    ],
                    providers: [
                        LBDComponentsService,
                        SerializerService,
                        BimViewerService,
                        ColorService,
                        VisibilityService
                    ],
                }]
        }] });

/*
 * Public API Surface of lbd-components
 */

/**
 * Generated bundle index. Do not edit.
 */

export { BimViewerComponent, BimViewerService, ColorService, ICDD, LBDComponentsModule, LBDComponentsService, ParsingProgress, SelectionMode, SerializerService, Tool, ViewerSettings, VisibilityService };
//# sourceMappingURL=ngx-lbd-components.mjs.map
